//-----------------------------------------------------------------------
//	SPARKLE V1.1
//	Inspired by Lft's Spindle and Krill's Loader
//	C64 Code
//	Tested on 1541-II, 1571, 1541 Ultimate-II+, and THCM's SX-64
//-----------------------------------------------------------------------
//	Memory Layout
//
//	0180	020e	Loader
//	01d9	01df	BitTab
//	01e0	01f0	IRQ Installer
//	020f	02e0	Depacker
//	02e1	02fb	Fallback IRQ
//	0300	03ff	Buffer
//
//-----------------------------------------------------------------------
//
//	Loader Call:		jsr $0180	Parameterless
//	IRQ Installer:		jsr $01e0	X/A = Subroutine Vector Lo/Hi
//					jsr $01e6	Without changing Subroutine Vector
//	Fallback IRQ:		    $02e1
//
//-----------------------------------------------------------------------

//Constants:
.const	Sp		=<$ff+$52	//#$51 - Spartan Stepping constant
.const	InvSp		=Sp^$ff	//#$ae

.const	ZP		=$02		//$02/$03

.const	busy		=$f8		//AO=1, CO=1, DO=1 on C64
.const	ready		=$08		//AO=1, CO=0, DO=0 on C64
.const	drivebusy	=$12		//AA=1, CO=0, DO=1 on Drive, CI=1, DI=0 on C64

.const	Buffer	=$0300

*=$0801	"Basic"		//Prg starts @ $0820, this gives the user
BasicUpstart (Start) 		//plenty of space to modify the basic start line

*=$0820	"Installer"

Start:	ldx	#<Cmd
		ldy	#>Cmd
		lda	#CmdEnd-Cmd
		jsr	$ffbd		//Filename = drive install code in command buffer

		lda	#$0f
		tay
		ldx	$ba
		bne	*+4
		ldx	#$08
		jsr	$ffba		//Logical parameters
		jsr	$ffc0		//Open vector

		sei
		lda	#$3c		// 0  0  1  1  1  1  0  0
		sta	$dd02		//DI|CI|DO|CO|AO|RS|VC|VC
		ldx	#$00		//Clear the lines
		stx	$dd00

		lda	#$35
		sta	$01

LCopyLoop:	lda	LoaderCode,x
		sta	$0180,x
		lda	LoaderCode+$80,x
		sta	$0200,x
		lda	LoaderCode+$100,x
		sta	$0280,x
		inx
		bpl	LCopyLoop
		
		ldx	#$7f
		txs			//Loader starts @ $180, so reduce stack to $100-$17f

		lda	#<NMI		//Install NMI vector
		sta	$fffa
		lda	#>NMI
		sta	$fffb

		lda	#busy		//=#$f8
		bit	$dd00		//Wait for "drive busy" signal	
		bmi	*-3
		sta	$dd00		//lock bus

		//First loader call

		lda	#>$10ad	//#>PrgStart-1	(Hi Byte)
		pha
		lda	#<$10ad	//#<PrgStart-1	(Lo Byte)
		pha
		jmp	Load		//Load first part, it may overwrite installer, so we use an alternative approach here

//----------------------------
Cmd:
//Load all 5 drive code blocks into buffers 0-4 at $300-$7ff on drive in one command!

.byte	'M','-','E',$05,$02	//-0204

		ldx	#$08		//-0206
		lda	#$12		//-0208	Sector
		tay			//-0209	Track
		sec			//-020a	Set Track and Sector number for buffers 04..00
		sty	$06,x		//-020c	Track = 18
		sta	$07,x		//-020e	Sectors 18,14,10,06,02
		sbc	#$04		//-0210	Interleave=4
		dex			//-0211
		dex			//-0212
		bpl	*-8		//-0214
		lda	#$04		//-0216	Load 5 blocks to buffers 04,03..00
		sta	$f9		//-0218	Buffer Pointer
		jsr	$d586		//-021b	Read Block into Buffer in Buffer Pointer 
		dec	$f9		//-021d	Decrease Buffer Pointer
		bpl	*-5		//-021f
					//		Maximize buffer utilization by preparing registers for Drive Code
		sei			//-0220
		lda	#$7a		//-0222	Set these $1800 bits to OUT
		ldy	#drivebusy	//-0224	CO=0, DO=1, AA=1
		jmp	$0300		//-0227	Execute Drive Code, X=#$00 after loading all 5 blocks (last buffer No=0) 
CmdEnd:

//----------------------------
//	C64 RESIDENT CODE
//	$0180-$02fb
//----------------------------

LoaderCode:

*=LoaderCode	"Loader"

.pseudopc $0180	{

Load:		lda	#$35		//Make sure RAM is turned off
		sta	$01
LastX:	ldx	#$00
		bne	LongMatch	//If LastX<>#$00, depack new part from buffer, otherwise, load next block
RcvBlock:	ldy	#ready	//Y=#$08, X=#$00
		sty	$dd00		//Clear CO and DO to signal Ready-To-Receive
		bit	$dd00		//Wait for Ready-To-Send from Drive
		bvs	*-3		//$dd00=#$4x - drive is busy, $0x - drive is ready	00,01
		stx	$dd00		//Release ATN							02-05
		dex			//Prepare X=#$ff for transfer					06,07
		jsr	Wait12	//Waste a few more cycles... (drive takes 16 cycles)	08-19

//-------------------------------------
//	    72-BYCLE RECEIVE LOOP
//		 Saves 15 bytes
//  Adds 2x10 cycles to the load time
//		  of one block
// (no detectable difference in speed)
//-------------------------------------

RcvLoop:
Read1:	lda	$dd00		//4		W1-W2 = 18 cycles on drive			20-23
		sty	$dd00		//4	8	Y=#$08 -> ATN=1
		lsr			//2	10
		lsr			//2	12
		inx			//2	14
		nop			//2	16
		ldy	#$00		//2	(18)

Read2:	ora	$dd00		//4		W2-W3 = 16 cycles
		sty	$dd00		//4	8	Y=#$00 -> ATN=0
		lsr			//2	10
		lsr			//2	12
SpComp:	cpx	#Sp		//2	14	Will be changed to #$ff in Spartan Step Delay
		beq	ChgJmp	//2/3	16/17 whith branch -----------|
		ldy	#$08		//2	(18/28)	ATN=1			|
					//						|
Read3:	ora	$dd00		//4		W3-W4 = 17 cycles		|
		sty	$dd00		//4	8	Y=#$08 -> ATN=1		|
		lsr			//2	10					|
		lsr			//2	12					|
		sta	LastBits+1	//4	16					|
		lda	#$c0		//2	(18)					|
					//						|
Read4:	and	$dd00		//4		W4-W1 = 16 cycles		|
		sta	$dd00		//4	8	A=#$X0 -> ATN=0		|
LastBits:	ora	#$00		//2	10					|
		sta	Buffer,x	//5	15					|
JmpRcv:	jmp	RcvLoop	//3	(18)	this could be BVC...	|
					//						|
//----------------------------						|
					//						|
ChgJmp:	ldy	#<SpSDelay	//2	19	<-----------------------|
		sty	JmpRcv+1	//4	23
		bne	Read3-2	//3	26	Branch always

//----------------------------
//		BITTAB
//----------------------------

//Literal Length BitTab values (tab entries 4/6/7) are corrected 
//so that the offset is always #$6d (= default offset of tab entry 7)

BitTab:		//$01d5							BitTab:		A after rol:
.byte	%10000001	//1 - Read 2nd bit					10000001	->	0000001x		go to 2/3

.byte	%00000000	//2 - Literal, no more bits (returns 1)		00000000	->	N/A			total: 2 bits
.byte	%10000010	//3 - Read 3rd bit					10000010	->	0000010x		go to 4/5 

.byte	%01100101	//4 - Literal - 4-5th bits (returns 2-5)		01100101	->	100101xx		total: 5 bits
.byte	%10000011	//5 - Read 4th bit					10000011	->	0000011x		go to 6/7

.byte	%00110011	//6 - Literal - 5-7th bits (returns 6-13)		00110011	->	10011xxx		total: 7 bits
.byte	%00001101	//7 - Literal - 5-9th bits (returns 14-45)	00001101	->	101xxxxx		total: 9 bits

.byte	$00		//4 bytes for future use...
.byte	$00
.byte	$00
.byte	$00

//----------------------------
//		IRQ INSTALLER
//		Call:	jsr $01e0
//		X/A=Player Lo/Hi
//----------------------------

InstallIRQ:	stx	IRQSub+1	//Installs a subroutine
		sta	IRQSub+2
		lda	#<IRQ		//Installs Fallback IRQ vector
		sta	$fffe
		lda	#>IRQ
		sta	$ffff
Wait12:		
Done:		rts

//----------------------------
//		LONG MATCH
//----------------------------

LongMatch:	clc			//C=0 NEEDED HERE for both branches!!! ALSO NEEDED FOR NEXT PART JUMP IF LastX<>#$00
		bne	NextFile	//A=#$3f - Next File in block (#$fc) - Also used as a trampoline for LastX jump (also needs C=0)
		dex			//A=#$3e - Long Match (#$f8), read next byte for Match Length (#$3e-#$fe)
		lda	Buffer,x	//If A=#$00 then this part is done, rest of the block in buffer is the beginning of the next part
		bne	MidConv	//Otherwise, converge with mid match (A=#$3e-#$fe here if branch taken)
		txa			//Branch not taken, transfer last X value to A
		bne	End		//ALWAYS, C=0, A=LastX<>#$00 here

//----------------------------
//		SPARTAN STEP DELAY
//----------------------------

SpSDelay:	lda	#<RcvLoop	//2	20	Restore Receive loop
		sta	JmpRcv+1	//4	24
		txa			//2	26
		eor	#InvSp	//2	28	Invert byte counter
		sta	SpComp+1	//4	32	SpComp+1=(#$2a <-> #$ff)
		bmi	RcvLoop	//3	(35) (Drive loop takes 33 cycles)

//----------------------------

		lda	#busy		//=#$f8
		sta	$dd00		//Bus lock

//------------------------------------------------------------
//		LZ DECOMPRESSION ROUTINE
//------------------------------------------------------------
//		$00->	- Bit Stream
//		($ff	- Block Count in first block only)
//		$ff	- Dest Address Lo
//		($fe	- IO Flag)
//		$fe	- Dest Address Hi
//		$fd->	- Byte Stream backwards (unpacked blocks are forward)
//------------------------------------------------------------

Depack:	ldx	#$00
		stx	BitCheck+1	//Reset BitCheck

		sec			//Token Bit, this keeps bit stream buffer<>#$00 until all 8 bits read
		rol	Buffer	//Move Compression Bit to C here
					//Only blocks occupied by a single file can be uncompressed

BlockCnt:	ldy	#$01		//Initial value of Block Counter = #$01, will decrease to #$00 after first block loaded
		dey
		bne	UpdateBCnt

		dex
		ldy	Buffer,x	//Last byte = BlockCnt, ByteStream = Buffer+$fe
UpdateBCnt:	sty	BlockCnt+1	//New value = total number of blocks occupied by this part-1
					//It is only used here to count blocks...
					//...to be able to properly read block address by skipping next block count value
					//End of a part is determined from the byte stream, not from here

NextFile:	dex			//Entry point for next file in block, C must be 0 here for subsequent files	
		lda	Buffer,x	//Lo Byte of Dest Address
		sta	ZP
		
		ldy	#$35		//Default value for $01, I/O=on
		dex
		lda	Buffer,x	//Hi Byte vs IO Flag=#$00
		bne	SkipIO
		dey			//Y=#$34, turn I/O off
		dex
		lda	Buffer,x	//This version can also load to zeropage!!!

SkipIO:	sta	ZP+1
		sty	$01		//Update $01
		dex

		lda	#%10000001	//Prepare value for first BitCheck		
		bcc	BitCheck	//Evaluate Compression Bit in C here

//----------------------------
//		UNPACKED
//----------------------------

		txa			//Uncompressed block is stored forward, not reversed
		tay			//I.e. address bytes point at beginning of uncompressed block[-1], not at end

UnPkdLoop:	lda	Buffer,y
		sta	(ZP),y	//(ZP)=Address-1
		dey			//Y=Max->#$01, #$00 not included
		bne	UnPkdLoop
		tya			//Unpacked block copy complete, A=#$00 to reset LastX

//----------------------------
//		END OF BLOCK
//----------------------------

End:		ldx	#$35		//A=Last X value if more date left in buffer or 0
		stx	$01
		sta	LastX+1	//Update LastX here, #$00 if block finished, otherwise Last X position		
		tax			//This is needed for bne to work
		bne	Done		//Last block complete, job complete
		jmp	RcvBlock	//Load next block, A=X=#$00 here

//----------------------------
//		MID MATCH
//----------------------------

MidMatch:	lda	Buffer,x	//C=0
		beq	End		//Match byte=#$00 -> end of block, will also reset LastX
		lsr
		lsr
		cmp	#$3e		//= (Long Match Tag/4)
		bcs	LongMatch	//Long Match (C=1) vs Mid Match (C=0)

MidConv:	tay			//Match Length=#$01-#$3d (mid) vs #$3e-#$fe (long)
		eor	#$ff
		adc	ZP		//C=0 here
		sta	ZP

		dex
		lda	Buffer,x	//Match Offset=$00-$ff+(C=1)=$01-$100

		bcs	ShortConv+1	//Skip sec
		dec	ZP+1
		bcc	ShortConv	//Converge with short match

//----------------------------
//		BITCHECK
//----------------------------

NextByte:	inc	BitCheck+1
BitCheck:	rol	Buffer
		beq	NextByte	//If buffer=#$00 then token bit is in C, go to next position in buffer
		rol
		bcc	BitCheck
		bmi	Literal	//A>=#$80, C=1 (bmi before beq assuming literals are more frequent than matches)
		beq	Match		//A=#$00, C=1

		clc
		tay			//Y=%10-%11, or %100-%111 here  
		lda	BitTab-1,y
		bne	BitCheck	//If A=#$00 - this is LitLen=#$01, otherwise go to BitCheck

//----------------------------
//		LITERALS
//----------------------------

		lda	#$94		//Lit0=#$94+#$6d+C(=0)=#$01 and C=1 after this addition
Literal:	adc	#$6d		//Offset is the same for all 4 literal lengths here
		sta	SubX+1	//Literal length, C=1 here ALWAYS
		tay
		eor	#$ff		//ZP=ZP+A^#$FF+C(=1) = ZP=ZP-A (e.g. A=#$0e -> ZP=ZP-0e)
		adc	ZP
		sta	ZP
		bcs	*+4
		dec	ZP+1

		txa
SubX:		axs	#$00		//X=X-(Literal+1) (e.g. Lit=#$00 -> X=X-(0+1))
		stx	LitCopy+1

		cpy	#$2d		//MaxLit+1=45
LitCopy:	lda	Buffer,y
		sta	(ZP),y
		dey
		bne	LitCopy
		bcs	ClrC		//if LitCnt=MaxLit - read 1 bit to see whether next sequence it Lit or Match

//----------------------------
//		SHORT MATCH
//----------------------------

Match:	lda	Buffer,x	//C=1 from BitCheck, C=0 from Literals
		anc	#$03		//Also clears C=0, needed after BitCheck
		beq	MidMatch	//C=0

ShortMatch:	tay			//Short Match Length=#$01-#$03
		eor	#$ff
		adc	ZP
		sta	ZP
		bcs	*+4
		dec	ZP+1

		lda	Buffer,x	//Short Match Offset=($00-$3f)+1=$01-$40
		lsr
		lsr
ShortConv:	sec
		adc	ZP
		sta	MatchCopy+1	//MatchCopy+1=ZP+(Buffer)+(C=1)
		lda	ZP+1
		adc	#$00
		sta	MatchCopy+2	//C=0 after this
		iny			//Y+=1 for bne to work
MatchCopy:	lda	$10ad,y	//Y=#$02-#$04 (short) vs #$02-#$3e (mid) vs #$3f-#$ff (long)
		sta	(ZP),y	//Y=#$00 is never used here - it is used as the End of Stream flag 
		dey
		bne	MatchCopy

		dex

ClrC:		clc			//NEEDED AFTER LITERALS!!!
		lda	#%10000000	//Read 1 bit
		bcc	BitCheck	//Branch ALWAYS

//----------------------------
//		FALLBACK IRQ
//----------------------------

IRQ:		pha			//$02de
		txa
		pha
		tya
		pha
		lda	$01
		pha
		lda	#$35
		sta	$01
IRQSub:	jsr	Done		//Music player or IRQ subroutine, installer @ $01e0
		inc	$d019
		pla
		sta	$01
		pla
		tay
		pla
		tax
		pla
NMI:		rti

//----------------------------

EndLoader:

.print "Loader Call:		" + toHexString(Load)
.print "BitTab:		" + toHexString(BitTab)
.print "IRQ Installer:	" + toHexString(InstallIRQ)
.print "Depack:		" + toHexString(Depack)
.print "Fallback IRQ:		" + toHexString(IRQ)
.print "Loader End:		" + toHexString(EndLoader-1)
}